#!/usr/bin/perl 
#  makelibspec.pl
#
#  Compose a ordered .spec of object modules to insert in a library
#
#  Usage: makelibspec.pl [-sd=<sourcedir>] [-td=<targetdir>] <libname>
#  output: <libname>.lib.spec
#  Uses the file <libname>.lib.def with the names of the object modules
#  to include in the library. 

#  The module names can be in any order;
#  this program rearranges the order of the modules in such a way that
#  the required modules of the library can be loaded by a single pass
#  over the library.
#
#  The file <libname>.lib.def is assumed in the <sourcedir>
#  The file <libname>.lib.spec will be stored in <sourcedir>
#  The object files are asuumed to be in the <targetdir>.
#  When either the <sourcedir> or <targetdir> are not specified,
#  the current directory is assumed.
#
#  Author: Theo Engel
#  Info:   Info@theoengel.nl
#
use strict;
use bytes;
my $DEBUG = 0;
use constant MAXNAMELENGTH => 10;   # max symbol name length

# SPECIAL MODULE: syslib, if defined, must always be the last module of a library
my $syslib = "ccinit.obj";
my $syslib_defined=0;

my @t;
my $sd=""; # source directory
my $td=""; # target directory
my $sdir="current";
my $tdir="current";
my $fi;

# argument 1
my $arg = shift (@ARGV);
if ($arg !~ /\-/ && $arg ne "")
{
	$fi=$arg;
}
elsif ($arg =~ /\-/)
{
	@t=split(/\=/,$arg);
	if($t[0] eq "-sd")
	{
		$sd=$t[1];
		$sdir=$sd;
	}
	elsif($t[0] eq "-td")
	{
		$td=$t[1];
		$tdir=$td;
	}
	else
	{
		print "Wrong 1st argument:$arg\n";
		print "Usage: makelibspec.pl [-sd=<sourcedir>] [-td=<targetdir>] <libname>\n";
		exit 1;
	}
	# argument 2
	$arg = shift (@ARGV);
	if ($arg !~ /\-/ && $arg ne "")
	{
		$fi = $arg;
	}
	elsif ($arg =~ /\-/)
	{
		@t=split(/\=/,$arg);
		if($t[0] eq "-sd")
		{
			$sd=$t[1];
			$sdir=$sd;			
		}
		elsif($t[0] eq "-td")
		{
			$td=$t[1];
			$tdir=$td;			
		}
		else
		{
			print "Wrong 2nd argument:$arg\n";
		    print "Usage: makelibspec.pl [-sd=<sourcedir>] [-td=<targetdir>] <libname>\n";			
			exit 1;
		}
		$arg = shift (@ARGV);
		if ($arg !~ /\-/ && $arg ne "")
		{
			$fi = $arg;
		}
		else
		{
			print "Wrong 3rd argument:$arg\n";
		    print "Usage: makelibspec.pl [-sd=<sourcedir>] [-td=<targetdir>] <libname>\n";			
			exit 1;
		}
	}
}
else
{
	print "No or wrong arguments\n";
	print "Usage: makelibspec.pl [-sd=<sourcedir>] [-td=<targetdir>] <libname>\n";	
	exit 1;
}



my $ldef  = $sd . '/' . $fi . '.lib.def';		  # library definition file
my $fspec = $sd . '/' . $fi . '.lib.spec';        # lib spec file (output)
my $fobj;                                         # module object file
my $fpfobj;                                       # module object file (incl full path)

print "Make library file $fspec according library definition $ldef\n";
print "Start reading $ldef\n";

open (IN,$ldef) || die "cannot open definition file $ldef: $!";
open (OUT, ">", $fspec) || die "cannot open outputfile $fspec: $!";

my $cin=0;		# number of module names read in from spec file
my $leader="false";
my $c =	0;	            # current char (8bit)
my $pc;		            # current char (7bit)
my $prevc;	            # previous char (8bit)

my $blty;

my $eom     = 0203;
my $null 	= 0;
my $som     = 0201;
my $mskp 	= 0177;     # 7 bit mask

my @mod;        # array with file name of the object modules
my %mextrns;    # hash with externals per module; key:<modfilename> value:<extrn1> <extern2> ...
my %entries;    # hash with all entry point; key:<entry> value:<modfilename>
my %mentries;   # hash with entries per module; key:<modfilename> value:<entry1> <entry2> ...(not required for the algoritm)
my $modname;    # object module's IDENT
my @extrns;
my $symbol;
my $end=0;
my $i;
my $j;
my $k;
my $l;
my $n;
my $temp;


# collect the object modules
while (<IN>)
{ 
	if($DEBUG) { print "$_";}
	$c=substr($_,0,1);
	if($DEBUG) {print "c:$c\n";}
	next if($c eq '#' || $c eq '*'); # skip comment lines
	if($_ =~ ':EOF')
	{
		print ":EOF\n";
		last;
	}
	chomp($_);
	$fobj= $_;
	$fpfobj = $td . '/' . $fobj;
	print "Get object module: $fpfobj\n";
	open (OBJ, $fpfobj) || die "cannot open objectfile $fpfobj: $!";
    # read the object module to collect its data
	if($syslib eq $_)
	{
		# ATTENTION: syslib in the library
		$syslib_defined = 1;
	}
	$mod[$cin] = $fobj;
	$cin=$cin+1;
    $end=0;	
	$leader="false";	
	while($end==0)
	{
		nextchar();
		#skip leader in case there
		if ($leader eq "false")
		{
			next if $c eq $null;
			$leader = "true";
			if($DEBUG) {print "Leader skipped\n";}
		}
		if($c == 0xFF)
		{
			# start of object module
			# collect (name,) entry points and externals of the current module			
			while($end==0)
			{
				nextchar();
				if($c == $som)
				{
					# IDENT	
					$modname = "";
					if($DEBUG) { print "IDENT ";}	
					$i=0;
					while($i<MAXNAMELENGTH)
					{
						nextchar();
						$modname = $modname . $pc;
						if($DEBUG) {print "$pc";}
						$i=$i+1;
					}
					if($DEBUG) {print "($cin-1)";}
					while(1)
					{
						nextchar();
						if($c eq $som)
						{
							$blty = nextchar();
							if($blty == 1)
							{
								# ENTRY
								$symbol = "";          # clear 	
								if($DEBUG) {print "ENTRY ";}
								$i=0;
								while($i<MAXNAMELENGTH)
								{
									nextchar();
									$symbol = $symbol . $pc;
									if($DEBUG) {print "$pc";}			
									$i=$i+1;
								}
								$symbol = rtrim($symbol);
								$mentries{$fobj}  = $mentries{$fobj} . $symbol . " ";	# enter entry point in hash
								$entries{$symbol} = $fobj;
							}
							elsif($blty == 4)
							{
								# EXTRN
								$symbol = "";          # clear 	
								if($DEBUG) {print "EXTRN ";}
								$i=0;
								while($i<MAXNAMELENGTH)
								{
									nextchar();
									$symbol = $symbol . $pc;
									if($DEBUG) {print "$pc";}
									$i=$i+1;
								}
								$symbol = rtrim($symbol);
								# insert external if not ther yet
								if($mextrns{$fobj} !~ /$symbol/)
								{
									$mextrns{$fobj} = $mextrns{$fobj} . $symbol . " ";	# enter external in hash
								}
							}
							elsif($blty == 7)
							{
								# END
								if($DEBUG) {print "END   ";}
								if($DEBUG) {print "\n";}
								# end of library module
								# module data is collected
								$end=1;
								last;
							}
						}
					}
				}
			}
		}
	}
	close(OBJ);
}
close(IN);


if($DEBUG)
{
	print "nbr modules:$cin modules:@mod\n";
	print "Collected Entry points\n";
    foreach $symbol (sort keys %entries)
	{
		print "E:$symbol M:$entries{$symbol}\n";
	}
	print "Per module, Externals, Entry Points\n";
	$i=0;
	while($i<($cin))
	{
		print "$mod[$i] X:$mextrns{$mod[$i]} E:$mentries{$mod[$i]}\n";
		$i=$i+1;
	}
}	

# sort the array @mod
if($DEBUG)
{
	print "Module ordering start\n";
	$k=0;
	while($k<$cin)
	{
		print "$mod[$k]\n";
		$k=$k+1;
	}
}

my $exch=1;
while($exch) # proceed until no exchanges any more
{
	$i=0;
	$exch=0;
	while($i<($cin) && $exch==0)
	{
		$j=0;	
		while($j<$cin && $exch==0)
		{
			# if extrn mod[j] is entry of mod[i] => exchange
			@extrns=split(/ /,$mextrns{$mod[$j]});      # externals of mod[j]
			$l=@extrns;                                 # number of externals
			$n=0;
			if($DEBUG) { print "i=$i j=$j l=$l modi=$mod[$i] modj=$mod[$j] XofModj:@extrns.\n";}
			while($n<$l && $exch==0)
			{
				$symbol=$extrns[$n];
				if($entries{$symbol} ne undef) # entry point in library ?
				{
					if($mod[$i] eq $entries{$symbol} && $i<$j)
					{
						$temp    = $mod[$i];
						$mod[$i] = $mod[$j];
						$mod[$j] = $temp;
						$exch=1; # restart
						if($DEBUG)
						{
							$k=0;
							if($DEBUG) { print "\n";} 
							while($k<$cin)
							{
								print "$mod[$k]\n";
								$k=$k+1;
							}
						}
					}
				}
				$n=$n+1;
			}
			$j=$j+1;
		}
		$i=$i+1;
	}
}

# generate .spec file
$i=0;
while($i<$cin)
{
	if($mod[$i] ne $syslib)
	{
		print OUT "* $mod[$i] X:$mextrns{$mod[$i]} E:$mentries{$mod[$i]}\n";
		print OUT "$mod[$i]\n";
	}
	$i=$i+1;
}
# syslib as the last module in the library
if($syslib_defined)
{
		print OUT "* $syslib X:$mextrns{$syslib} E:$mentries{$syslib}\n";
		print OUT "$syslib\n";
}
print OUT ":EOF\n";
print "Library Spec file ready\n";
close(OUT);
exit;


sub nextchar {
	    if(read(OBJ,$c,1) == 0)
		{
			return -1;						#EOF
		}
		$prevc = $c; 				        # save previous char
		$c = unpack("C", $c); 	            # get new current char
		$pc= $c & $mskp;
		$pc= chr($pc);				        # current char in printable form
		return ($c & 0xFF);
}


sub rtrim {
	my $s = shift;
	my $l = length($s);
	my $i = $l;
	if ($l > 0)
	{
		while($i > 0)
		{
			if((substr($s,$i-1,1)) ne " ")
			{ 
				last;
			}
			else
			{
				$i=$i-1;
			}
		}
	}
	else
	{
		return "";
	}
	if($i == 0)
	{
		return "";
	}
	else
	{
		return substr($s,0,$i);
	}
}



